import protoattend.utils as utils


def test_get_default_configuration():
    default_config = utils.load_default_config()
    assert "candidate_batch_size" in default_config


def test_get_optimizer_names():
    assert "AdaBound" in utils.get_available_optimizers()



from protoattend.module import ProtoAttendModule
from protoattend import utils
from typing import Union, Tuple, Dict
import numpy as np
import torch
import torch.utils.data as tud
import pytorch_lightning as pl
from argparse import Namespace
import logging

logger = logging.getLogger(__name__)


class ProtoAttend:
    def __init__(self, encoder, num_classes, model_output_dim: int = None, **hparams):
        self.hparams = utils.load_default_config()
        self.hparams.update(hparams)
        if not isinstance(self.hparams, Namespace):
            self.hparams = utils.wrap_namespace(self.hparams)

        self.hparams.num_classes = num_classes

        self.hparams.d_encoder = model_output_dim
        if not model_output_dim:
            self.hparams.d_encoder = utils.infer_model_output_dim(encoder)

        self.module = ProtoAttendModule(encoder, self.hparams)

    @staticmethod
    def make_dataset(
        X_train: Union[np.ndarray, torch.Tensor, tud.Dataset, tud.DataLoader],
        X_val: Union[np.ndarray, torch.Tensor, tud.Dataset, tud.DataLoader],
        y_train: Union[np.ndarray, torch.Tensor] = None,
        y_val: Union[np.ndarray, torch.Tensor] = None,
    ) -> Tuple[tud.Dataset, tud.Dataset]:
        if isinstance(X_train, tud.Dataset):
            assert isinstance(
                X_val, tud.Dataset
            ), f"If X_train is a tud.Dataset, X_val must be a tud.Dataset as well!"
            return X_train, X_val
        return tud.TensorDataset(X_train, y_train), tud.TensorDataset(X_val, y_val)

    def process_dataloaders(
        self, train_dataloader: tud.DataLoader, val_dataloader: tud.DataLoader
    ) -> Tuple[tud.DataLoader, tud.DataLoader]:
        """
        Given a pair of DataLoaders, it adds `candidate_batch_size` onto the
        current batch size of each DataLoader. Also, if train and val batch
        sizes are different, sets them the same and gives a warning. Finally,
        sets module config to appropriately split batches during training.
        Parameters
        ----------
        train_dataloader: the tud.DataLoader for train set.
        val_dataloader: the tud.DataLoader for val set.

        Returns
        -------
        train_dataloader: modified tud.DataLoader for train set.
        val_dataloader: modified tud.DataLoader for val set.
        """
        train_dataloader_kwargs = {
            k: v for k, v in train_dataloader.__dict__.items() if not k.startswith("_")
        }
        train_dataloader_kwargs.pop("batch_sampler")
        val_dataloader_kwargs = {
            k: v for k, v in val_dataloader.__dict__.items() if not k.startswith("_")
        }
        val_dataloader_kwargs.pop("batch_sampler")

        cand_bs = self.hparams.candidate_batch_size
        self.module.hparams_to_init.input_ratio = utils.compute_input_candidate_ratio(
            cand_bs, train_dataloader.batch_size
        )

        if train_dataloader.batch_size != val_dataloader.batch_size:
            logger.warning(
                f"Setting val_dataloader `batch_size` equal to train_dataloader. This is required for ProtoAttend."
            )
        new_batch_size = train_dataloader.batch_size + cand_bs
        train_dataloader_kwargs.update({"batch_size": new_batch_size})
        val_dataloader_kwargs.update({"batch_size": new_batch_size})

        train_loader = tud.DataLoader(**train_dataloader_kwargs)
        val_loader = tud.DataLoader(**val_dataloader_kwargs)
        return train_loader, val_loader

    def make_dataloaders(
        self,
        X_train: Union[np.ndarray, torch.Tensor, tud.Dataset, tud.DataLoader],
        X_val: Union[np.ndarray, torch.Tensor, tud.Dataset, tud.DataLoader],
        y_train: Union[np.ndarray, torch.Tensor] = None,
        y_val: Union[np.ndarray, torch.Tensor] = None,
    ) -> Tuple[tud.DataLoader, tud.DataLoader]:
        if not isinstance(X_train, tud.DataLoader):
            assert not isinstance(
                X_val, tud.DataLoader
            ), f"If X_train is a dataset, X_val must be a dataset as well!"

            train_dataset, val_dataset = self.make_dataset(
                X_train, X_val, y_train, y_val
            )
            train_dataloader_params = self.hparams["train_dataloader_params"]
            val_dataloader_params = self.hparams["val_dataloader_params"]
            train_dataloader = tud.DataLoader(train_dataset, **train_dataloader_params)
            val_dataloader = tud.DataLoader(val_dataset, **val_dataloader_params)
        else:
            train_dataloader = X_train
            val_dataloader = X_val
        return self.process_dataloaders(train_dataloader, val_dataloader)

    def fit(
        self,
        X_train: Union[np.ndarray, torch.Tensor, tud.Dataset, tud.DataLoader],
        X_val: Union[np.ndarray, torch.Tensor, tud.Dataset, tud.DataLoader],
        y_train: Union[np.ndarray, torch.Tensor] = None,
        y_val: Union[np.ndarray, torch.Tensor] = None,
    ):

        train_dataloader, val_dataloader = self.make_dataloaders(
            X_train, X_val, y_train, y_val
        )
        self.module.train_loader = train_dataloader
        self.module.val_loader = val_dataloader

        trainer_params = self.hparams.trainer_params.__dict__
        trainer = pl.Trainer(**trainer_params)
        trainer.fit(self.module)

    def predict(self, X_test):
        # test_loader = self.make_dataloaders(X_test)
        test_loader = X_test
        out = self.module.infer(test_loader)
        return out



import yaml
import os.path as osp
from pathlib import Path
import torch.nn as nn
import torch
import torch_optimizer
from functools import singledispatch
from types import SimpleNamespace


def is_capitalized(string: str) -> bool:
    return string[0].isupper()


def get_available_optimizers():
    native_optims = [optim for optim in dir(torch.optim) if is_capitalized(optim)]
    extra_optims = [optim for optim in torch_optimizer.__all__ if is_capitalized(optim)]
    all_optims = native_optims + extra_optims
    all_optims.sort()
    return all_optims


def compute_input_candidate_ratio(input_batch_size, candidate_batch_size):
    return input_batch_size / (input_batch_size + candidate_batch_size)


def load_default_config():
    with open(Path(__file__).parent / "default_configuration.yaml", "r") as file:
        return yaml.load(file, Loader=yaml.FullLoader)


def infer_model_output_dim(model):
    modules = [child for child in model.children() if hasattr(child, "weight")]
    final_module = modules[-1]
    if not isinstance(final_module, nn.Linear):
        raise ValueError(
            f"Could not infer model output dimensions. "
            f"Please provide these dimensions manually by passing "
            f"in `model_output_dim` arg in `ProtoAttend`"
        )
    return final_module.out_features


@singledispatch
def wrap_namespace(ob):
    return ob


@wrap_namespace.register(dict)
def _wrap_dict(ob):
    return SimpleNamespace(**{k: wrap_namespace(v) for k, v in ob.items()})


@wrap_namespace.register(list)
def _wrap_list(ob):
    return [wrap_namespace(v) for v in ob]



import torch
import torch.nn as nn
import torch.utils.data as tud
import pytorch_lightning as pl
import torch_optimizer
from sparsemax import Sparsemax
from tqdm import tqdm
import math
import copy
import protoattend.utils as utils
from typing import Union, List, Dict, Iterable
import logging

logger = logging.getLogger(__name__)

# Types
LOG_TYPE = Dict[str, torch.Tensor]


class ProtoAttendModule(pl.LightningModule):
    def __init__(self, encoder, hparams):
        super().__init__()
        self._is_cached = False

        self.database_keys = None
        self.database_values = None
        self.database_labels = None

        self.train_loader = None
        self.val_loader = None

        self.hparams_to_init = copy.deepcopy(hparams)  # Used to initialize module

        self.hparams = hparams
        self.hparams.lr = hparams.optimizer_params.lr
        self.lr = self.hparams_to_init.optimizer_params.lr

        self.d_encoder = self.hparams_to_init.d_encoder
        self.d_attention = self.hparams_to_init.d_attention
        self.sqrt_d_att = math.sqrt(self.d_attention)
        self.d_val = self.hparams_to_init.d_val
        self.d_intermediate = self.hparams_to_init.d_intermediate

        self.encoder = encoder
        self.encoder_and_ffn = nn.Sequential(
            encoder,
            nn.Linear(self.d_encoder, self.d_intermediate),
            nn.ReLU(inplace=True),
            nn.LayerNorm(self.d_intermediate),
        )
        self.Q_inp = nn.Sequential(
            nn.Linear(self.d_intermediate, self.d_attention), nn.ReLU(inplace=True)
        )
        self.K_cand = nn.Sequential(
            nn.Linear(self.d_intermediate, self.d_attention), nn.ReLU(inplace=True)
        )
        self.V = nn.Sequential(
            nn.Linear(self.d_intermediate, self.d_val),
            nn.ReLU(inplace=True),
            nn.LayerNorm(self.d_val),
        )

        self.sparsemax = Sparsemax(-1)
        self.softmax = nn.Softmax(-1)

        self.classifier = nn.Linear(self.d_val, self.hparams_to_init.num_classes)

    def attention(self, Q, K, sparse=True):
        """Computes attention logits, then pass through sparsemax or softmax"""
        attention_logits = Q.mm(K.t()) / self.sqrt_d_att
        if sparse:
            return self.sparsemax(attention_logits)
        return self.softmax(attention_logits)

    def attention_sparsity_loss(self, attention):
        eps = self.hparams_to_init.sparsity_epsilon
        entropy = attention * (attention + eps).log()
        entropy = entropy.sum(-1).mean()
        return -entropy

    @staticmethod
    def confidence_loss(attention, input_y, candidate_y, reduce=True):
        """Maximise attention weights where input == candidate labels"""
        indicator = input_y.unsqueeze(-1).eq(candidate_y.unsqueeze(0))  # B_img x B_cand
        confidence = (attention * indicator).sum(-1)
        if reduce:
            return -confidence.mean()
        return -confidence

    def convex_superposition(self, p, q, alpha: Union[List[float], float] = 0.5):
        if isinstance(alpha, list):
            return [self.classifier(a * p + (1 - a) * q) for a in alpha]
        return self.classifier(alpha * p + (1 - alpha) * q)

    def forward(self, batch):
        (
            (input_x, input_y),
            (candidate_x, candidate_y),
        ) = self.split_batch_into_input_and_candidate(batch)

        input_x = input_x.to(next(self.encoder.parameters()))
        candidate_x = candidate_x.to(next(self.encoder.parameters()))

        input_enc = self.encoder_and_ffn(input_x)
        cand_enc = self.encoder_and_ffn(candidate_x)

        # Project encodings onto query, key and value spaces
        Q_inp = self.Q_inp(input_enc)
        K_cand = self.K_cand(cand_enc)
        V_inp = self.V(input_enc)
        V_cand = self.V(cand_enc)

        # Compute attention probabilities and weighted prototype embedding
        attention = self.attention(Q_inp, K_cand)
        weighted_prototype_embedding = attention.mm(V_cand)

        # Compute convex superpositions
        self_logits, intermediate_logits, prototype_logits = self.convex_superposition(
            V_inp, weighted_prototype_embedding, alpha=[0.0, 0.5, 1.0]
        )

        output = {"intermediate_logits": intermediate_logits, "labels": input_y}

        if input_y is not None:
            # Compute loss components
            ce_loss = nn.CrossEntropyLoss()

            # Self loss: Normal supervised loss
            self_loss = ce_loss(self_logits, input_y)

            # Intermediate loss: Equal mix of supervised and prototype loss
            intermediate_loss = ce_loss(intermediate_logits, input_y)

            # Prototype loss: Use logits only from weighted sum of prototype candidates
            prototype_loss = ce_loss(prototype_logits, input_y)

            # Compute confidence loss
            confidence_loss = self.confidence_loss(attention, input_y, candidate_y)

            # Compute sparsity loss
            sparsity_loss = self.attention_sparsity_loss(attention)

            # Compute overall loss
            loss = (
                self_loss
                + intermediate_loss
                + prototype_loss
                + self.hparams_to_init.sparsity_weight * sparsity_loss
                + self.hparams_to_init.confidence_weight * confidence_loss
            )

            output.update(
                {
                    "loss": loss,
                    "self_loss": self_loss,
                    "intermediate_loss": intermediate_loss,
                    "prototype_loss": prototype_loss,
                    "confidence_loss": confidence_loss,
                    "sparsity_loss": sparsity_loss,
                }
            )

        return output

    def training_step(self, batch, batch_idx):
        output = self.forward(batch)
        output.pop("intermediate_logits")
        output.pop("labels")

        tensorboard_logs = {"loss": output["loss"]}
        output.update({"log": tensorboard_logs})
        return output

    def validation_step(self, batch, batch_idx):
        output = self.forward(batch)
        predictions = output.pop("intermediate_logits").argmax(-1)
        correct = (predictions == output.pop("labels")).float().mean()

        # Replace all keys in output dictionary; eg "loss" -> "val_loss"
        renamed_keys_output = {}
        for old_key, value in output.items():
            new_key = "val_" + old_key
            renamed_keys_output[new_key] = output[old_key]
        renamed_keys_output.update({"correct": correct})
        return renamed_keys_output

    def validation_epoch_end(self, output: List[LOG_TYPE]):
        output: Dict[str, Union[torch.Tensor, LOG_TYPE]] = self.collate(output)
        tensorboard_logs = dict()
        for key, value in output.items():
            tensorboard_logs[key] = value
        output.update({"log": tensorboard_logs})

        return output

    def train_dataloader(self):
        return self.train_loader

    def val_dataloader(self):
        return self.val_loader

    @staticmethod
    def get_dataloader_without_shuffle(loader: tud.DataLoader):
        train_dataloader_kwargs = {
            k: v for k, v in loader.__dict__.items() if not k.startswith("_")
        }
        train_dataloader_kwargs.pop("batch_sampler")
        train_dataloader_kwargs.pop("sampler")
        train_dataloader_kwargs.update({"shuffle": False})
        return tud.DataLoader(**train_dataloader_kwargs)

    def cache_database_computations(self):
        """Computes keys, values and stores labels for the database."""
        logger.info(f"Caching database key and values")

        # Initialize temporary state - disable dataloader shuffle, gradients, dropout, batchNorm
        train_loader = self.get_dataloader_without_shuffle(self.train_loader)
        torch.set_grad_enabled(False)
        self.eval()

        outputs = []
        for idx, batch in tqdm(enumerate(train_loader)):
            x, y = batch
            x = x.to(next(self.encoder.parameters()))
            y = y.to(next(self.encoder.parameters()))
            encoding = self.encoder_and_ffn(x)
            encoded_key = self.K_cand(encoding)
            encoded_value = self.V(encoding)
            output = {"keys": encoded_key, "values": encoded_value, "labels": y}
            outputs.append(output)

        outputs = self.collate(outputs)
        self.database_keys = outputs["keys"]
        self.database_values = outputs["values"]
        self.database_labels = outputs["labels"]

        self._is_cached = True

        # Restore original state
        self.train()
        torch.set_grad_enabled(True)
        # TODO: sort this out
        # setattr(self.train_loader, "shuffle", original_shuffle_bool)

    def infer(self, loader):
        """Runs inference on a tud.DataLoader."""

        if not self._is_cached:
            self.cache_database_computations()

        # Setup initial state
        torch.set_grad_enabled(False)
        self.eval()

        outputs = []
        for idx, batch in tqdm(enumerate(loader)):
            x, *_ = batch
            x = x.to(next(self.encoder.parameters()))

            encoding = self.encoder_and_ffn(x)
            queries = self.Q_inp(encoding)
            values = self.V(encoding)

            attention = self.attention(queries, self.database_keys)
            weighted_prototype_embedding = attention.mm(self.database_values)

            intermediate_logits = self.convex_superposition(
                values, weighted_prototype_embedding, alpha=0.5
            )
            predictions = intermediate_logits.argmax(-1)
            confidence = -self.confidence_loss(
                attention, predictions, self.database_labels, reduce=False
            )
            # attn: B_input x B_database (e.g 16 x 30_000)
            attn_weights, attn_indices = attention.topk(5, dim=-1)
            output = {
                "prototype_idxs": attn_indices.cpu(),
                "prototype_weights": attn_weights.cpu(),
                "confidence": confidence.cpu(),
                "predictions": predictions.cpu(),
            }
            outputs.append(output)

        outputs = self.collate(outputs)

        # Restore back to initial state
        self.train()
        torch.set_grad_enabled(True)

        return outputs

    def split_batch_into_input_and_candidate(self, batch):
        """Splits a batch into input and candidate groups."""
        x, y = batch

        batch_size = x.size(0)
        cutoff = math.ceil(batch_size * self.hparams_to_init.input_ratio)
        rand = torch.randperm(batch_size, device=x.device)
        input_idxs = rand[:cutoff]
        cand_idxs = rand[cutoff:]

        input_x = x.index_select(0, input_idxs)
        input_y = y.index_select(0, input_idxs)
        candidate_x = x.index_select(0, cand_idxs)
        candidate_y = y.index_select(0, cand_idxs)

        return (input_x, input_y), (candidate_x, candidate_y)

    def configure_optimizers(self):
        """Configures optimizers and LR schedulers"""
        optim_name = self.hparams_to_init.optimizer
        if hasattr(torch.optim, optim_name):
            optimizer_class = getattr(torch.optim, optim_name)
        elif hasattr(torch_optimizer, optim_name):
            optimizer_class = getattr(torch_optimizer, optim_name)
        else:
            raise AttributeError(
                f"Optimizer: {optim_name} is not available. "
                f"Available optimizers: {utils.get_available_optimizers()}"
            )
        optimizer_params = copy.deepcopy(self.hparams_to_init.optimizer_params.__dict__)
        optimizer_params.pop("lr")
        optimizer = optimizer_class(self.parameters(), self.hparams.lr, **optimizer_params)

        to_return = [optimizer]

        if hasattr(self.hparams_to_init, "scheduler"):
            scheduler_params = self.hparams_to_init.scheduler_params.__dict__
            scheduler = {
                "monitor": scheduler_params.pop("monitor", "val_loss"),
                "interval": scheduler_params.pop("interval", "epoch"),
                "frequency": scheduler_params.pop("frequency", 1),
            }
            scheduler_class = getattr(
                torch.optim.lr_scheduler, self.hparams_to_init.scheduler
            )
            scheduler.update(
                {"scheduler": scheduler_class(optimizer, **scheduler_params)}
            )
            to_return = ([optimizer], [scheduler])
        return to_return

    @staticmethod
    def collate(output: List[LOG_TYPE]) -> LOG_TYPE:
        """Collates output from many validation steps"""
        collated_output = {}
        example = output[0]
        for key in example:
            tensor_dim = example[key].dim()
            if tensor_dim == 0:
                collated_output[key] = torch.stack([x[key] for x in output]).mean()
            elif tensor_dim == 1 or tensor_dim == 2:
                collated_output[key] = torch.cat([x[key] for x in output])
            else:
                raise ValueError(f"Unexpected tensor dimension: {tensor_dim}")
        return collated_output



"""Top-level package for protoattend."""
# fmt: off
__author__ = """Laksh Aithani"""
__email__ = 'lakshaithanii@gmail.com'
__version__ = '0.1.13'
# fmt: on


from protoattend.module import ProtoAttendModule
from protoattend.protoattend import ProtoAttend
